from typing import Any
from django.urls import path, re_path, URLPattern
from django.conf import settings


def dynamicpath(rel_path: str, view: Any, kwargs: dict[str, Any]| None  = None, name: str | None = None) -> URLPattern:
    if settings.MY_SCRIPT_NAME:
        return path(settings.MY_SCRIPT_NAME.lstrip("/") + "/" + rel_path,  view, kwargs, name)
    else:
        return path(rel_path, view, kwargs, name)



def re_dynamicpath(regex_path: str, view: Any, kwargs: dict[str, Any] | None = None, name: str | None = None) -> URLPattern:
    if settings.MY_SCRIPT_NAME:
        return re_path(rf"^{settings.MY_SCRIPT_NAME.lstrip('/')}/" + regex_path.lstrip("^"), view, kwargs, name)
    else:
        return re_path(regex_path, view, kwargs, name)




import json
from typing import Any, Dict, Optional
from channels.testing import ApplicationCommunicator


class GraphQLHttpTestClient:
    """
    A simple test client for executing GraphQL queries via HTTP using Django Channels' ApplicationCommunicator.
    
    Example:
        client = GraphQLHttpTestClient(application)
        result = await client.execute("{ __typename }")
    """

    def __init__(
        self,
        application: Any,
        path: str = "/graphql",
        headers: Optional[Dict[str, str]] = None
    ) -> None:
        """
        Initialize the test client.

        Args:
            application: The ASGI application to test.
            path: The HTTP path to send the GraphQL request to (default is "/graphql").
            headers: Optional dictionary of additional HTTP headers.
        """
        self.application = application
        self.path = path
        self.headers = headers or {}

    async def execute(
        self,
        query: str,
        variables: Optional[Dict[str, Any]] = None,
        operation_name: Optional[str] = None,
        timeout: int = 5
    ) -> Dict[str, Any]:
        """
        Execute a GraphQL query or mutation via HTTP.

        Args:
            query: The GraphQL query string.
            variables: Optional dictionary of GraphQL variables.
            operation_name: Optional name of the operation (if query contains multiple).
            timeout: Maximum time to wait for a response in seconds.

        Returns:
            A dictionary containing the JSON-decoded GraphQL response.

        Raises:
            AssertionError: If the HTTP response is not 200 OK or malformed.
        """
        body: Dict[str, Any] = {"query": query}
        if variables is not None:
            body["variables"] = variables
        if operation_name is not None:
            body["operationName"] = operation_name

        encoded_body = json.dumps(body).encode()

        # Prepare headers for ASGI
        header_list = [(b"content-type", b"application/json")]
        for key, value in self.headers.items():
            header_list.append((key.encode(), value.encode()))

        communicator = ApplicationCommunicator(self.application, {
            "type": "http",
            "method": "POST",
            "path": self.path,
            "headers": header_list,
            "body": encoded_body,
        })

        await communicator.send_input({
            "type": "http.request",
            "body": encoded_body,
        })

        response_start = await communicator.receive_output(timeout=timeout)
        assert response_start["type"] == "http.response.start", f"Unexpected type: {response_start}"
        assert response_start["status"] == 200, f"Unexpected status: {response_start['status']}"

        response_body = await communicator.receive_output(timeout=timeout)
        assert response_body["type"] == "http.response.body", f"Unexpected type: {response_body}"

        return json.loads(response_body["body"])




class HttpGetTestClient:
    """
    A minimal ASGI test client for performing raw HTTP GET requests using ApplicationCommunicator.

    Example:
        client = HttpGetTestClient(application)
        response_text = await client.get("/schema")
    """

    def __init__(
        self,
        application: Any,
        headers: Optional[Dict[str, str]] = None
    ) -> None:
        self.application = application
        self.headers = headers or {}

    async def get(
        self,
        path: str,
        accept: str = "text/plain",
        timeout: int = 5
    ) -> str:
        """
        Perform a GET request to the given path.

        Args:
            path: HTTP path to request (e.g., "/schema").
            accept: MIME type for the Accept header.
            timeout: Maximum time to wait for a response.

        Returns:
            Decoded response body as a string.

        Raises:
            AssertionError: If response type or status is unexpected.
        """
        header_list: List[Tuple[bytes, bytes]] = [(b"accept", accept.encode())]
        for k, v in self.headers.items():
            header_list.append((k.encode(), v.encode()))

        communicator = ApplicationCommunicator(self.application, {
            "type": "http",
            "method": "GET",
            "path": path,
            "headers": header_list,
            "body": b"",
        })

        await communicator.send_input({"type": "http.request", "body": b""})

        response_start = await communicator.receive_output(timeout=timeout)
        assert response_start["type"] == "http.response.start", f"Unexpected type: {response_start}"
        assert response_start["status"] == 200, f"Unexpected status: {response_start['status']}"

        response_body = await communicator.receive_output(timeout=timeout)
        assert response_body["type"] == "http.response.body", f"Unexpected type: {response_body}"

        return response_body["body"].decode()



from strawberry.channels import ChannelsConsumer, ChannelsRequest
from dataclasses import dataclass
from typing import Any, Dict, Literal, Mapping, Optional, Protocol


class User(Protocol):
    id: int
    sub: str

    def is_anonymous(self) -> bool:
        """Check if the user is anonymous."""
        return self.id is None


class Client(Protocol):
    id: str
    client_id: str
    
    
class Organization(Protocol):
    id: int
    slug: str


@dataclass
class UniversalRequest:
    _extensions: Dict[str, Any]
    _client: Optional[Client] = None
    _user: Optional[User] = None
    _organization: Optional[Organization] = None

    @property
    def user(self) -> User:
        """Get the user associated with the request."""
        if self._user is None:
            raise ValueError(
                "User is not set in the request. Do you have a strawberry extension setting this?"
            )

        return self._user

    @property
    def client(self) -> Client:
        if self._client is None:
            raise ValueError(
                "Client is not set in the request.  Do you have a strawberry extension setting this?"
            )

        return self._client

    def set_user(self, user: User) -> None:
        """Set an extension value in the request."""
        self._user = user
        
    def set_organization(self, organization: Organization) -> None:
        """Set an organization in the request."""
        self._organization = organization

    def set_client(self, client: Client) -> None:
        """Set an extension value in the request."""
        self._client = client

    def get_extension(self, name: str) -> Any:
        """Get an extension value from the request."""

        if name not in self._extensions:
            raise ValueError(f"Extension {name} is not set in the request.")
        return self._extensions.get(name)

    def set_extension(self, name: str, value: Any) -> None:
        """Set an extension value in the request."""

        self._extensions[name] = value


@dataclass
class WsContext:
    request: UniversalRequest
    connection_params: Dict[str, Any]
    consumer: ChannelsConsumer
    extensions: Optional[Dict[str, Any]] = None
    type: Literal["ws"] = "ws"


@dataclass
class HttpContext:
    request: UniversalRequest
    headers: Mapping[str, str]
    type: Literal["http"] = "http"


Context = HttpContext | WsContext



from asgiref.typing import (
    ASGIApplication,
    Scope,
    ASGIReceiveCallable,
    ASGISendCallable,
    ASGISendEvent,
)
from typing import Awaitable


class CorsMiddleware:
    def __init__(self, app: ASGIApplication):
        self.app = app

    async def __call__(
        self, scope: Scope, receive: ASGIReceiveCallable, send: ASGISendCallable
    ) -> None:
        if scope["type"] == "http" and scope["method"] == "OPTIONS":
            # preflight request. reply successfully:
            headers = [
                (b"Access-Control-Allow-Origin", b"*"),
                (b"Access-Control-Allow-Headers", b"Authorization, Content-Type"),
                (b"Access-Control-Allow-Methods", b"GET, POST, PUT, DELETE, OPTIONS"),
                (b"Access-Control-Max-Age", b"86400"),
            ]
            await send(
                {"type": "http.response.start", "status": 200, "headers": headers, "trailers": False}
            )
            await send(
                {
                    "type": "http.response.body",
                    "body": b"",
                    "more_body": False,
                }
            )
        else:

            async def wrapped_send(event: ASGISendEvent) -> None:
                if event["type"] == "http.response.start":
                    original_headers = event.get("headers") or []
                    access_control_allow_origin = b"*"

                    if access_control_allow_origin is not None:
                        # Construct a new event with new headers
                        event = {
                            "type": "http.response.start",
                            "status": event["status"],
                            "headers": [
                                p
                                for p in original_headers
                                if p[0] != b"access-control-allow-origin"
                            ]
                            + [
                                (
                                    b"access-control-allow-origin",
                                    access_control_allow_origin,
                                ),
                            ],
                            "trailers": event.get("trailers", False),
                        }

                await send(event)

            await self.app(scope, receive, wrapped_send)

        return None
